/* CSA.cpp
   Copyright 2003, V. Makinen, all rights reserved.

   This file contains an implementation of the Compact Suffix Array
   index structure [1,2,3].

   This software may be used freely for any non-commercial purpose. However, when distributed,
   the original source must be clearly stated, and, when the source code is
   distributed, the copyright notice must be retained and any alterations in
   the code must be clearly marked. No warranty is given regarding the quality
   of this software.

   [1] Veli Mkinen. Compact Suffix Array. In Proc. 11th Annual Symposium on
       Combinatorial Pattern Matching (CPM 2000), Springer-Verlag LNCS 1848,
       pp. 305-319, Montral, June 2000.
   [2] Veli Mkinen. Trade off Between Compression and Search Times in Compact
       Suffix Array. In Proc. 3rd Workshop on Algorithm Engineering and
       Experimentation (ALENEX 2001), Springer-Verlag LNCS 2153, pp. 189-201,
       Washington D.C., January 2001.
   [3] Veli Mkinen. Compact Suffix Array - A Space-Efficient Full-Text Index.
       Fundamenta Infomaticae, Special Issue - Computing Patterns in Strings, 
       56(1-2):191-210,2003.
*/

//---------------------------------------------------------------------------
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/times.h>
#include <assert.h>
#include <string.h>
#include "qsufsort.c"
#define maxx(A,B) ((A) > (B) ? (A) : (B))
#define minn(A,B) ((A) < (B) ? (A) : (B))
//---------------------------------------------------------------------------

int hzz=HZ;

static struct tms startown;

void startownclock()
{
	times(&startown);
}

long stopownclock()
{
	struct tms t;

	times(&t);
	return(t.tms_utime - startown.tms_utime);
}


unsigned n,m;
unsigned w=32;
bool print=true;
int report=1;
long tuc, maxtucindex=0, maxtuc=0, maxtucsize=0;
unsigned maxdepth, maxLCP;
double sumdepth;
unsigned *depthcount = (unsigned *)malloc(10000*sizeof(unsigned));
unsigned operations;
unsigned C=15,D=10;
unsigned bitsn=0;
unsigned savemem=0;

typedef struct {
   unsigned signbit : 1;
   unsigned LCP : 31;
} lcpmask;

typedef struct {
   unsigned signbit : 1;
   unsigned pos : 31;
} posmask;

typedef struct {
   unsigned jump : 16;
   unsigned depth : 16;
} jumpmask;

typedef union {
   lcpmask as_lcp;
   jumpmask as_jump;
} lcpjumpmask;

typedef struct SAElement {
    posmask a;
} TSAElement;

typedef struct LCPElement {
  lcpjumpmask a;
} TLCPElement;

typedef struct {
   unsigned int lLCP;
   unsigned int rLCP;
} TLCPPair;

class TSA {
public:
    unsigned posbits;
    unsigned lcpbits;
    unsigned *pos;
    unsigned *lcp;
    unsigned *B;    
    bool jumpinnext;
private:
    void SetField(unsigned *A, unsigned len, unsigned index, unsigned x)
      {unsigned i=index*len/32, 
                j=index*len-i*32;
       unsigned mask = (j+len < 32 ? ~0u << j+len : 0) 
                       | (32-j < 32 ? ~0u >> (32-j) : 0);
       A[i] = (A[i] & mask) | x << j;
       if (j+len>32) {
          mask = ((~0u) << (len+j-32));  
          A[i+1] = (A[i+1] & mask)| x >> (32-j);
       } 
      }; 
    unsigned GetField(unsigned *A, unsigned len, unsigned index)         
      {unsigned i=index*len/32,
                j=index*len-32*i,
                result;                
       if (j+len <= 32)
          result = (A[i] << (32-j-len)) >> (32-len);
       else {
          result = A[i] >> j;
          result = result | (A[i+1] << (32-(j+len-32))) >> (32-len);
       }
       return result;
      };

public:
    void printbits(unsigned x) 
      {unsigned k; for (k=0; k<32; k++) printf("%d",x<<(32-k-1)>>31);
      printf("\n");};
    void CreatePosArray(unsigned bits, unsigned n) 
      {posbits = bits; pos = (unsigned *)malloc((n*bits)/8+1);
       CreateBooleanArray(n);};
    void CreateLCPArray(unsigned bits, unsigned n, bool jump)
      {lcpbits = bits; lcp = (unsigned *)malloc((n*bits)/8+1);
      assert(lcp!=NULL); jumpinnext = jump;};    
    void CreateBooleanArray(unsigned n)
      {B = (unsigned *)malloc(n/8+1);};    
    void DeletePosArray() {free(pos);};
    void DeleteLCPArray() {free(lcp);};
    void DeleteBooleanArray(unsigned n) {free(B);};        
    void SetLCP(unsigned index,unsigned x)   
      {x = x | ((GetField(lcp,lcpbits,index) >> (lcpbits-1)) << (lcpbits-1));
       SetField(lcp,lcpbits,index,x);};
    void SetLCPl(unsigned index)   
      {SetField(lcp,lcpbits,index,
       (1u<<(lcpbits-1)) | GetField(lcp,lcpbits,index));};
    void SetLCPr(unsigned index)   
      {SetField(lcp,lcpbits,index,
       GetField(lcp,lcpbits,index) & ~(1u<<(lcpbits-1)));};
    void SetJump(unsigned index, unsigned x)
      {if (!jumpinnext) SetField(lcp,lcpbits/2,2*index,x);
       else SetField(lcp,lcpbits,index+1,x);};
    void SetDepth(unsigned index, unsigned x)
      {if (!jumpinnext) SetField(lcp,lcpbits/2,2*index+1,x);
       else SetField(lcp,lcpbits,index,x);};
    void SetPos(unsigned index, unsigned x)
      {SetField(pos,posbits,index,x);};
    void SetB(unsigned index, unsigned x)
      {SetField(B,1,index,x);};
    unsigned GetLCP(unsigned index)   
      {return GetField(lcp,lcpbits,index) & ~(1<<(lcpbits-1));};
    unsigned GetLCPl(unsigned index)   
      {return GetField(lcp,lcpbits,index)>>(lcpbits-1);};
    unsigned GetLCPr(unsigned index)   
      {return GetField(lcp,lcpbits,index)>>(lcpbits-1)==0;};
    unsigned GetJump(unsigned index)
      {if (!jumpinnext) return GetField(lcp,lcpbits/2,2*index);
       else return GetField(lcp,lcpbits,index+1);};    
    unsigned GetDepth(unsigned index)
      {if (!jumpinnext) return GetField(lcp,lcpbits/2,2*index+1);
       else return GetField(lcp,lcpbits,index);};
    unsigned GetPos(unsigned index)
      {return GetField(pos,posbits,index);};
    bool GetB(unsigned index)
      {return GetField(B,1,index);};
};


unsigned SuffixArrayCompactSize = 0;
char* text;
char* pattern;
char* filename;
TSA *SA,*CSA;

void MakeSuffixArray(int alg);
void PrintSuffixArray(TSA *SArray,unsigned length);
void PrintSuffixArray(TSA *SArray, unsigned length);
unsigned SearchFromSA(char *pattern, TSA *SArray, unsigned length);
char *SubString(char * A,unsigned start, unsigned count);
void UnCompactLink(TSA *CArray, unsigned Cindex, unsigned depth,
                   TSA *Area, unsigned
                   &Areasize,unsigned remaining, unsigned startparent);


char *SubString(char * A,unsigned start, unsigned count)
{
   char * Result = (char*)malloc((count+1)*sizeof(char));
   unsigned i;
   i=start;
   unsigned j=i;
   while (i < start+count && A[i] != '\0') {
      if (A[i]!= '\n') {
         Result[j-start] = A[i];
         j++;
      }
      i++;
   }
   Result[i-start] = '\0';
   return Result;
}

int ComparePos(unsigned pos, char *pattern)
{
   unsigned j=0;
   while (pos+j < n && j < m && text[pos+j]==pattern[j])
      j++;
   if (j >= m)
      return 0;
   else if (pos+j >= n)
      return +1;
   else if (text[pos+j]>pattern[j])
      return -1;
   else
      return +1;

}

int LCP(char *a, int apos, char *b, int bpos)
{
   int i=0;
   while (a[apos+i]!='\0' && b[bpos+i] != '\0' && a[apos+i]==b[bpos+i])
       i++;
   return i;
}


inline int CompareSuffixes(char *S, unsigned x, unsigned y, unsigned depth)
{
      int inc=0;
      if (x==y) return 0;
      while (x+inc+depth < n && y+inc+depth < n && S[x + inc+depth] == S[y + inc+depth])
         inc++;
      if (x+inc+depth == n)
         return -(inc+1);
      else if (y+inc+depth == n)
         return +(inc+1);
      else
         return S[x+inc+depth] < S[y+inc+depth] ? -(inc+1) : +inc+1;
}

void CalcAdjacentLCPsSA(TSA *SArray)
{
   unsigned i,j;
   unsigned index,x;
   unsigned comparisons=0;
   for (i=0; i<n; i++) {
      assert(SArray->GetPos(i)<n);
      SArray->SetLCP(SArray->GetPos(i),i);
   }
   unsigned lastLCP=0, newLCP=0;
   double sumLCP=0.0, sumLCPv=0.0;
   for (i=0; i<n; i++) {
      index = SArray->GetLCP(i);
      if (index == 0)
         SArray->SetLCP(i,0);
      else {
         if (lastLCP > 0)
            lastLCP--;
         newLCP = 
         abs(CompareSuffixes(text,SArray->GetPos(index),SArray->GetPos(index-1),
                lastLCP));
         SArray->SetLCP(i,lastLCP+newLCP-1);
         comparisons += newLCP;
      }
      lastLCP = SArray->GetLCP(i);
   }
   maxLCP =0;
   for (i=0; i<n; i++) {
      if (SArray->GetLCP(SArray->GetPos(i)) >maxLCP)
         maxLCP = SArray->GetLCP(SArray->GetPos(i));
      sumLCP += SArray->GetLCP(SArray->GetPos(i));
   }
   for (i=0; i<n; i++) {
      sumLCPv += (SArray->GetLCP(SArray->GetPos(i))-sumLCP/n)*(SArray->GetLCP(SArray->GetPos(i))-sumLCP/n);
   }
   unsigned pLCP,cLCP;
   for (i=0; i<n; i++)
      SArray->SetB(i,0);
   for (i=0; i<n; i++) {
      j = SArray->GetPos(i);
      x = i;
      while (SArray->GetB(j)==0) {
         index = j;
         j = SArray->GetPos(index);
         SArray->SetPos(index,x);
         SArray->SetB(index,1);
         x = index;
      }
   }
   for (i=0; i<n; i++)
      SArray->SetB(i,0);
   for (i=0; i<n; i++) {
      j = SArray->GetPos(i);
      x = i;
      pLCP = SArray->GetLCP(x);
      while (SArray->GetB(j)==0) {
         index = j;
         j = SArray->GetPos(index);
         cLCP = SArray->GetLCP(index);
         SArray->SetLCP(index,pLCP);
         pLCP = cLCP;
         SArray->SetPos(index,x);
         SArray->SetB(index,1);
         x = index;
      }
   }
   for (i=0; i<n; i++)
      SArray->SetB(i,0);
   if (print) {
      printf("comparisons=%d <= 2*n=%d\n",comparisons,2*n);
      printf("maxLCP=%d\n",maxLCP);
      printf("avgLCP=%f\n",sumLCP/n);
      printf("sqrt(variance)=%f\n",sqrt(sumLCPv/n));
      fflush(stdout);
   }
}


TLCPPair CalcLCPsFromAdjacentLCPs(TSA *CArray,unsigned l, unsigned r)
{
   unsigned m = (l+r) / 2;
   unsigned i;
   TLCPPair lp,rp,tp;
   tp.lLCP = 0;
   tp.rLCP = 0;
   if (CArray->GetB(m))
      m--;
   assert(!CArray->GetB(m));
   if (m==l+1)
      tp.lLCP = CArray->GetLCP(m);
   else if (CArray->GetB((l+m)/2) && (l+m)/2 <= l+1) {
      tp.lLCP = CArray->GetLCP(m);
      for (i=l+1;i<m;i++)
         if (!CArray->GetB(i) && CArray->GetLCP(i) < tp.lLCP)
            tp.lLCP = CArray->GetLCP(i);
   }
   else {
      lp = CalcLCPsFromAdjacentLCPs(CArray,l,m);
      tp.lLCP = minn(lp.lLCP,lp.rLCP);
   }
   if (m==r-1)
      tp.rLCP = CArray->GetLCP(r);
   else if (CArray->GetB((m+r)/2) && (m+r)/2 <= m+1) {
      tp.rLCP = CArray->GetLCP(r);
      for (i=m+1;i<r;i++)
         if (!CArray->GetB(i) && CArray->GetLCP(i) < tp.rLCP)
            tp.rLCP = CArray->GetLCP(i);
   }
   else {
      rp = CalcLCPsFromAdjacentLCPs(CArray,m,r);
      tp.rLCP = minn(rp.lLCP,rp.rLCP);
   }
   if (tp.lLCP >= tp.rLCP) {
      CArray->SetLCPl(m);
      CArray->SetLCP(m,tp.lLCP);
   }
   else {
      CArray->SetLCPr(m);
      CArray->SetLCP(m,tp.rLCP);
   }
   if (l==0 && CArray->GetLCPl(0)==0) {
      CArray->SetLCP(0,tp.lLCP);
      CArray->SetLCPr(0);
      CArray->SetLCP(r,tp.rLCP);
      CArray->SetLCPl(r);
   }
   return tp;
}

TLCPPair CalcLCPs(TSA *CArray,unsigned l, unsigned r)
{
   unsigned m = (l+r) / 2;
   unsigned i;
   TLCPPair lp,rp,tp;
   tp.lLCP = 0;
   tp.rLCP = 0;
   if (CArray->GetB(m))
      m--;
   assert(!CArray->GetB(m));
   if (m==l+1 || (CArray->GetB((l+m)/2) && (l+m)/2 <= l+1)) {
      tp.lLCP =
      abs(CompareSuffixes(text,CArray->GetPos(l),CArray->GetPos(m),0))-1;
   }
   else {
      lp = CalcLCPs(CArray,l,m);
      tp.lLCP = minn(lp.lLCP,lp.rLCP);
   }
   if (m==r-1 ||(CArray->GetB((m+r)/2) && (m+r)/2 <= m+1)) {
      tp.rLCP =
      abs(CompareSuffixes(text,CArray->GetPos(m),CArray->GetPos(r),0))-1;
   }
   else {
      rp = CalcLCPs(CArray,m,r);
      tp.rLCP = minn(rp.lLCP,rp.rLCP);
   }
   if (tp.lLCP >= tp.rLCP) {
      CArray->SetLCPl(m);
      CArray->SetLCP(m,minn(tp.lLCP,maxLCP));
   }
   else {
      CArray->SetLCPr(m);
      CArray->SetLCP(m,minn(tp.rLCP,maxLCP));
   }
   if (l==0 && CArray->GetLCPl(0)==0) {
      CArray->SetLCP(0,minn(tp.lLCP,maxLCP));
      CArray->SetLCPr(0);
      CArray->SetLCP(r,minn(tp.rLCP,maxLCP));
      CArray->SetLCPl(r);
   }
   return tp;
}

unsigned SearchFromCSA(char *pattern, TSA *SArray, unsigned length)
{
   unsigned k;
   unsigned L=0;
   unsigned R=length-1;
   unsigned i=(L+R)/2;
   unsigned Lraja=0;
   unsigned Rraja=0;
   unsigned matches=0;
   unsigned limit;
   if (SArray->GetB(i))
      i--;
   // Search for the left boundary
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Lraja=0;
   else {
      while (L<R-1 && i!=L) {
         if (ComparePos(SArray->GetPos(i),pattern) <= 0)
            R = i;
         else
            L = i;
         i = (R+L)/2;
         if (SArray->GetB(i))
            i--;
      }
      if (i==L) {
         Lraja=L+1;
         if (L==R-3 && ComparePos(SArray->GetPos(R-1),pattern) > 0)
            Lraja = R;
      }
      else
         Lraja=R;
   }
   // Search for the right boundary
   L=0;
   R=length-1;
   i=(L+R)/2;
   if (SArray->GetB(i))
      i--;
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Rraja=0;
   else {
      while (L < R-1 && i!=L) {
         if (ComparePos(SArray->GetPos(i),pattern) >= 0)
            L = i;
         else
            R = i;
         i = (R+L)/2;
         if (SArray->GetB(i))
            i--;
      }
      if (i==L) {
         Rraja=L+1;
         if (L== R-3 && ComparePos(SArray->GetPos(R-1),pattern) >= 0)
            Rraja = R-1;
      }
      else
         Rraja = L;
   }
   if (Lraja+1 < Rraja && !report) {
      if (print) printf("Pattern (%s) found\n",pattern);
      return 1;
   }   
   else if ((Lraja==Rraja && SArray->GetB(Lraja)) || Lraja < Rraja) {
      TSA *DArray = new TSA();
      DArray->CreatePosArray(SArray->posbits,(Rraja-Lraja+1)*C);
      unsigned linkstart=0;
      unsigned j=0;
      for (i=Lraja; i<=Rraja; i++) {
         if (SArray->GetB(i)) {
            linkstart=0;
            UnCompactLink(SArray,
                       i,
                       0,
                       DArray,
                       j,
                       SArray->GetDepth(i),
                       0);
         }
         else {
            DArray->SetPos(j++,SArray->GetPos(i));
         }
      }
      matches=0;
      j--;
      for (i=0; i<=j; i++)
         if (ComparePos(DArray->GetPos(i),pattern)==0)
            break;
      for (; j>i; j--)
         if (ComparePos(DArray->GetPos(j),pattern)==0)
            break;
      for (i; i<=j; i++) {
         if (print && report)
            printf("Match (%s) at %d\n",pattern,DArray->GetPos(i));
         else if (!report) {
            matches = 1;
            if (print)
               printf("Pattern (%s) found\n",pattern);
            break;          
	 }
         matches++;
      }
      DArray->DeletePosArray();
      delete DArray;
      if (matches == 0 && !report && print)
         printf("Pattern (%s) not found\n",pattern);
      return matches;
   }
   else if (Lraja==Rraja) {
      // an occurrence found
      if (ComparePos(SArray->GetPos(Lraja),pattern)==0) {
         if (print)
            printf("Match (%s) at %d\n",pattern,SArray->GetPos(Lraja));
         return 1;
      }
   }
   if (!report && print)
         printf("Pattern (%s) not found\n",pattern);
   return 0;
}

unsigned SearchFromCSACount(char *pattern, TSA *SArray, unsigned length)
{
   unsigned k;
   unsigned L=0;
   unsigned R=length-1;
   unsigned i=(L+R)/2;
   unsigned Lraja=0;
   unsigned Rraja=0;
   unsigned matches=0;
   unsigned limit;
   if (SArray->GetB(i))
      i--;
   // Search for the left boundary
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Lraja=0;
   else {
      while (L<R-1 && i!=L) {
         if (ComparePos(SArray->GetPos(i),pattern) <= 0)
            R = i;
         else
            L = i;
         i = (R+L)/2;
         if (SArray->GetB(i))
            i--;
      }
      if (i==L) {
         Lraja=L+1;
         if (L==R-3 && ComparePos(SArray->GetPos(R-1),pattern) > 0)
            Lraja = R;
      }
      else
         Lraja=R;
   }
   // Search for the right boundary
   L=0;
   R=length-1;
   i=(L+R)/2;
   if (SArray->GetB(i))
      i--;
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Rraja=0;
   else {
      while (L < R-1 && i!=L) {
         if (ComparePos(SArray->GetPos(i),pattern) >= 0)
            L = i;
         else
            R = i;
         i = (R+L)/2;
         if (SArray->GetB(i))
            i--;
      }
      if (i==L) {
         Rraja=L+1;
         if (L== R-3 && ComparePos(SArray->GetPos(R-1),pattern) >= 0)
            Rraja = R-1;
      }
      else
         Rraja = L;
   }
   if ((Lraja==Rraja && SArray->GetB(Lraja)) || Lraja < Rraja) {
      TSA *DArray = new TSA();
      DArray->CreatePosArray(SArray->posbits,C);
      unsigned j=0;
      if (SArray->GetB(Lraja)) {
         UnCompactLink(SArray,
                       Lraja,
                       0,
                       DArray,
                       j,
                       SArray->GetDepth(Lraja),
                       0);
	 // search for the start of the occurrences	      
         j--;
         for (i=0; i<=j; i++)
            if (ComparePos(DArray->GetPos(i),pattern)==0)
               break;
	 matches += SArray->GetDepth(Lraja)-i;	 
      } else matches++;	 
      for (i=Lraja+1;i<Rraja;i++)
         if (SArray->GetB(i))
	    matches += SArray->GetDepth(i);
	 else 
	    matches++;
      j = 0;
      if (SArray->GetB(Rraja)) {
         UnCompactLink(SArray,
                       Rraja,
                       0,
                       DArray,
                       j,
                       SArray->GetDepth(Rraja),
                       0);
	 // search for the end of the occurrences	      
         j--;
         for (i=j; i!=0; i--)
            if (ComparePos(DArray->GetPos(i),pattern)==0)
               break;
	 matches += i+1;	 
      } else matches++;	 
      DArray->DeletePosArray();
      delete DArray;
      return matches;
   }
   else if (Lraja==Rraja) {
      // an occurrence found
      if (ComparePos(SArray->GetPos(Lraja),pattern)==0) {
         matches = 1;
      }      
   }
   return 0;
}

unsigned SearchFromSA(char *pattern, TSA *SArray, unsigned length)
{
   unsigned L=0;
   unsigned R=length-1;
   unsigned i=(L+R)/2;
   unsigned Lraja=0;
   unsigned Rraja=0;
   unsigned matches=0;
   // Search for the left boundary
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Lraja=0;
   else {
      while (L<R-1) {
         if (ComparePos(SArray->GetPos(i),pattern) <= 0)
            R = i;
         else
            L = i;
         i = (R+L)/2;
      }
      Lraja=R;
   }
   // Search for the right boundary
   L=0;
   R=length-1;
   i=(L+R)/2;
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Rraja=0;
   else {
      while (L < R-1) {
         if (ComparePos(SArray->GetPos(i),pattern) >= 0)
            L = i;
         else
            R = i;
         i = (R+L)/2;
      }
      Rraja=L;
   }
   matches=0;
   if (!report) {
      if (Lraja <= Rraja) {
         if (print) printf("Pattern (%s) found\n",pattern);
         return 1;
      }
      else {
         if (print) printf("Pattern (%s) not found\n",pattern);
         return 0;
      }
   }
   for (i=Lraja; i<= Rraja; i++) {
      matches++;
      if (print)
         printf("Match (%s) at %d\n",pattern,SArray->GetPos(i));
   }
   return matches;
}

unsigned SearchFromSACount(char *pattern, TSA *SArray, unsigned length)
{
   unsigned L=0;
   unsigned R=length-1;
   unsigned i=(L+R)/2;
   unsigned Lraja=0;
   unsigned Rraja=0;
   unsigned matches=0;
   // Search for the left boundary
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Lraja=0;
   else {
      while (L<R-1) {
         if (ComparePos(SArray->GetPos(i),pattern) <= 0)
            R = i;
         else
            L = i;
         i = (R+L)/2;
      }
      Lraja=R;
   }
   // Search for the right boundary
   L=0;
   R=length-1;
   i=(L+R)/2;
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Rraja=0;
   else {
      while (L < R-1) {
         if (ComparePos(SArray->GetPos(i),pattern) >= 0)
            L = i;
         else
            R = i;
         i = (R+L)/2;
      }
      Rraja=L;
   }
   return maxx(0,(int)(Rraja-Lraja+1));
}


unsigned SearchFromCSAWithLCP(char *pattern, TSA *SArray, unsigned length)
{
   unsigned k;
   unsigned L=0;
   unsigned R=length-1;
   unsigned i=(L+R)/2;
   unsigned Lraja=0;
   unsigned Rraja=0;
   unsigned matches=0;
   unsigned limit;
   unsigned lLCP, rLCP;
   unsigned mm=0;
   unsigned LCPparent,lLCPthis,rLCPthis;
   lLCP = LCP(text,SArray->GetPos(L),pattern,0);
   rLCP = LCP(text,SArray->GetPos(R),pattern,0);
   if (SArray->GetLCP(L) > SArray->GetLCP(R))
      LCPparent = SArray->GetLCP(L);
   else
      LCPparent = SArray->GetLCP(R);
   if (SArray->GetB(i))
      i--;
   assert(!SArray->GetB(i));
   // Search for the left boundary
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Lraja=0;
   else {
      while (R>L+1 && i!=L) {
         if (SArray->GetLCPl(i)) {
            rLCPthis = LCPparent;
            lLCPthis = SArray->GetLCP(i);
         }
         else {
            lLCPthis = LCPparent;
            rLCPthis = SArray->GetLCP(i);
         }
         if (lLCP >= rLCP)
            if (lLCPthis >= lLCP)
               mm = lLCP+LCP(text,SArray->GetPos(i)+lLCP,pattern,lLCP);
            else
               mm = lLCPthis;
         else
            if (rLCPthis >= rLCP)
               mm = rLCP+LCP(text,SArray->GetPos(i)+rLCP,pattern,rLCP);
            else
               mm = rLCPthis;
         if (mm==m || pattern[mm] <= text[SArray->GetPos(i)+mm]) {
            R = i;
            rLCP = mm;
            LCPparent = lLCPthis;
         }
         else {
            L = i;
            lLCP = mm;
            LCPparent = rLCPthis;
         }
         i = (R+L)/2;
         if (SArray->GetB(i))
            i--;
      }
      if (i==L) {
         Lraja=L+1;
         if (L==R-3 && ComparePos(SArray->GetPos(R-1),pattern) > 0)
            Lraja = R;
      }
      else
         Lraja=R;
   }
   // Search for the right boundary
   L=0;
   R=length-1;
   i=(L+R)/2;
   lLCP = LCP(text,SArray->GetPos(L),pattern,0);
   rLCP = LCP(text,SArray->GetPos(R),pattern,0);
   if (SArray->GetLCP(L) > SArray->GetLCP(R))
      LCPparent = SArray->GetLCP(L);
   else
      LCPparent = SArray->GetLCP(R);
   if (SArray->GetB(i))
      i--;
   assert(!SArray->GetB(i));
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Rraja=0;
   else {
      while (L < R-1 && i!=L) {
         if (SArray->GetLCPl(i)) {
            rLCPthis = LCPparent;
            lLCPthis = SArray->GetLCP(i);
         }
         else {
            lLCPthis = LCPparent;
            rLCPthis = SArray->GetLCP(i);
         }
         if (lLCP >= rLCP)
            if (lLCPthis >= lLCP)
               mm = lLCP+LCP(text,SArray->GetPos(i)+lLCP,pattern,lLCP);
            else
               mm = lLCPthis;
         else
            if (rLCPthis >= rLCP)
               mm = rLCP+LCP(text,SArray->GetPos(i)+rLCP,pattern,rLCP);
            else
               mm = rLCPthis;
         if (mm==m || pattern[mm] >= text[SArray->GetPos(i)+mm]) {
            L = i;
            lLCP = mm;
            LCPparent = rLCPthis;
         }
         else {
            R = i;
            rLCP = mm;
            LCPparent = lLCPthis;
         }
         i = (R+L)/2;
         if (SArray->GetB(i))
            i--;
      }
      if (i==L) {
         Rraja=L+1;
         if (L== R-3 && ComparePos(SArray->GetPos(R-1),pattern) >= 0)
            Rraja = R-1;
      }
      else
         Rraja = L;
   }
   if (Lraja < Rraja && !report) {
      if (print) printf("Pattern (%s) found\n",pattern);
      return 1;
   }
   else if ((Lraja==Rraja && SArray->GetB(Lraja)) || Lraja < Rraja) {
      TSA *DArray = new TSA();
      DArray->CreatePosArray(SArray->posbits,(Rraja-Lraja+1)*C);
      unsigned linkstart=0;
      unsigned j=0;
      for (i=Lraja; i<=Rraja; i++) {
         if (SArray->GetB(i)) {
            linkstart=0;
            UnCompactLink(SArray,
                       i,
                       0,
                       DArray,
                       j,
                       SArray->GetDepth(i),
                       0);
         }
         else {
            DArray->SetPos(j++,SArray->GetPos(i));
         }
      }
      matches=0;
      j--;
      for (i=0; i<=j; i++)
         if (ComparePos(DArray->GetPos(i),pattern)==0)
            break;
      for (; j>i; j--)
         if (ComparePos(DArray->GetPos(j),pattern)==0)
            break;
      for (i; i<=j; i++) {
         if (print && report)
            printf("Match (%s) at %d\n",pattern,DArray->GetPos(i));
         else if (!report) {
            matches = 1;
            if (print)
               printf("Pattern (%s) found\n",pattern);
            break;          }
         matches++;
      }
      delete DArray;
      if (matches == 0 && !report && print)
         printf("Pattern (%s) not found\n",pattern);
      return matches;
   }
   else if (Lraja==Rraja) {
      // an occurrence found
      if (ComparePos(SArray->GetPos(Lraja),pattern)==0) {
         if (print)
            printf("Match (%s) at %d\n",pattern,SArray->GetPos(Lraja));
         return 1;
      }
   }
   if (!report && print)
         printf("Pattern (%s) not found\n",pattern);
   return 0;
}


unsigned SearchFromSAWithLCP(char *pattern, TSA *SArray,unsigned length)
{
   unsigned L=0;
   unsigned R=length-1;
   unsigned i=(L+R)/2;
   unsigned Lraja=0;
   unsigned Rraja=0;
   unsigned matches=0;
   unsigned lLCP, rLCP;
   unsigned mm=0;
   unsigned LCPparent;
   unsigned lLCPthis;
   unsigned rLCPthis;
   lLCP = LCP(text,SArray->GetPos(L),pattern,0);
   rLCP = LCP(text,SArray->GetPos(R),pattern,0);
   if (SArray->GetLCP(L) > SArray->GetLCP(R))
      LCPparent = SArray->GetLCP(L);
   else
      LCPparent = SArray->GetLCP(R);
   // Search for the left boundary
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Lraja=0;
   else {
      while (R>L+1 && i!=L) {
         if (SArray->GetLCPl(i)) {
            rLCPthis = LCPparent;
            lLCPthis = SArray->GetLCP(i);
         }
         else {
            lLCPthis = LCPparent;
            rLCPthis = SArray->GetLCP(i);
         }
         if (lLCP >= rLCP)
            if (lLCPthis >= lLCP)
               mm = lLCP+LCP(text,SArray->GetPos(i)+lLCP,pattern,lLCP);
            else
               mm = lLCPthis;
         else
            if (rLCPthis >= rLCP)
               mm = rLCP+LCP(text,SArray->GetPos(i)+rLCP,pattern,rLCP);
            else
               mm = rLCPthis;
         if (mm==m || pattern[mm] <= text[SArray->GetPos(i)+mm]) {
            R = i;
            rLCP = mm;
            LCPparent = lLCPthis;
         }
         else {
            L = i;
            lLCP = mm;
            LCPparent = rLCPthis;
         }
         i = (R+L)/2;
      }
      Lraja=R;
   }
   // Search for the right boundary
   L=0;
   R=length-1;
   i=(L+R)/2;
   lLCP = LCP(text,SArray->GetPos(L),pattern,0);
   rLCP = LCP(text,SArray->GetPos(R),pattern,0);
   if (SArray->GetLCP(L) > SArray->GetLCP(R))
      LCPparent = SArray->GetLCP(L);
   else
      LCPparent = SArray->GetLCP(R);
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Rraja=0;
   else {
      while (L < R-1 && i!=L) {
         if (SArray->GetLCPl(i)) {
            rLCPthis = LCPparent;
            lLCPthis = SArray->GetLCP(i);
         }
         else {
            lLCPthis = LCPparent;
            rLCPthis = SArray->GetLCP(i);
         }
         if (lLCP >= rLCP)
            if (lLCPthis >= lLCP)
               mm = lLCP+LCP(text,SArray->GetPos(i)+lLCP,pattern,lLCP);
            else
               mm = lLCPthis;
         else
            if (rLCPthis >= rLCP)
               mm = rLCP+LCP(text,SArray->GetPos(i)+rLCP,pattern,rLCP);
            else
               mm = rLCPthis;
         if (mm==m || pattern[mm] >= text[SArray->GetPos(i)+mm]) {
            L = i;
            lLCP = mm;
            LCPparent = rLCPthis;
         }
         else {
            R = i;
            rLCP = mm;
            LCPparent = lLCPthis;
         }
         i = (R+L)/2;
      }
         Rraja = L;
   }
   if (!report) {
      if (Lraja <= Rraja) {
         if (print) printf("Pattern (%s) found\n",pattern);
         return 1;
      }
      else {
         if (print) printf("Pattern (%s) not found\n",pattern);
         return 0;
      }
   }
   for (i=Lraja; i<= Rraja; i++) {
      matches++;
      if (print)
         printf("Match (%s) at %d\n",pattern,SArray->GetPos(i));
   }
   return matches;
}

unsigned MarkLinksToSuffixArray(TSA *SArray, TSA *DArray)
{
   unsigned newsize=0;
   unsigned i=0;
   unsigned j;
   unsigned x;
   unsigned char *pathdepth=(unsigned char *)malloc(n*sizeof(unsigned char));
   unsigned index;
   unsigned minLCP;
   newsize=n;
   for (i=0; i<n; i++) {
      DArray->SetPos(SArray->GetPos(i),i);
      DArray->SetB(i,0);
      SArray->SetB(i,0);
      pathdepth[i]=0;
   }
   for (i=0; i<n; i++) {
      x = DArray->GetPos(i);
      DArray->SetPos(i,SArray->GetPos(i));
      SArray->SetPos(i,x);
   }
   i = 0;
   for (i=1; i<n-2; i++) {
      if (DArray->GetPos(i) == 0)
         x = n+1;
      else
         x = SArray->GetPos(DArray->GetPos(i)-1);
      if (x < n-1 &&
          DArray->GetPos(i+1) > 0 &&
          SArray->GetPos(DArray->GetPos(i+1)-1) == x+1 &&
          (D > 255 || (pathdepth[i] + pathdepth[x] < D &&
          pathdepth[i+1] + pathdepth[x+1] < D))) {
         if (D < 256 && pathdepth[x] == 0) {
            pathdepth[x] = pathdepth[i]+1;
            pathdepth[SArray->GetPos(DArray->GetPos(i)-1+pathdepth[x])] = pathdepth[x];
            if (pathdepth[x+1]+pathdepth[i+1]+1>D) {
	       pathdepth[x]=0;
               pathdepth[SArray->GetPos(DArray->GetPos(i)-1+pathdepth[x])] =
	       pathdepth[i];
	       i++;
	       continue;
	    }
         }
         else if (D < 256) {
            pathdepth[SArray->GetPos(DArray->GetPos(i)-1-pathdepth[x])] =
              pathdepth[x]+ pathdepth[i] + 1;
            pathdepth[SArray->GetPos(DArray->GetPos(i)+pathdepth[i])] =
              pathdepth[x]+ pathdepth[i] + 1;
            if (pathdepth[x+1]+pathdepth[i+1]+1>D) {
	       pathdepth[SArray->GetPos(DArray->GetPos(i)-1-pathdepth[x])]=pathdepth[x];
               pathdepth[SArray->GetPos(DArray->GetPos(i)+pathdepth[i])] = pathdepth[i];
	       i++;
	       continue;
	    }
         }
         DArray->SetB(i,1);
         DArray->SetPos(i,x);
         if (D < 256 && pathdepth[x+1] == 0) {
            pathdepth[x+1] = pathdepth[i+1]+1;
            pathdepth[SArray->GetPos(DArray->GetPos(i+1)-1+pathdepth[x+1])] =
            pathdepth[x+1];
         }
         else if (D<256) {
            pathdepth[SArray->GetPos(DArray->GetPos(i+1)-1-pathdepth[x+1])] =
              pathdepth[x+1]+ pathdepth[i+1] + 1;
            pathdepth[SArray->GetPos(DArray->GetPos(i+1)+pathdepth[i+1])] =
              pathdepth[x+1]+ pathdepth[i+1] + 1;
         }
         DArray->SetB(i+1,1);
         DArray->SetPos(i+1,x+1);
         newsize--;
         j = 1;
         while (j+1 < C && x+j+1 <n && i+j+1 <n-1 &&
                (D>255 || (pathdepth[i+j+1] + pathdepth[x+j+1] < D)) &&
                DArray->GetPos(i+j+1) > 0 &&
                SArray->GetPos(DArray->GetPos(i+j+1)-1) == x+j+1) {
            if (D <256 && pathdepth[x+j+1] == 0) {
               pathdepth[x+j+1] = pathdepth[i+j+1]+1;
               pathdepth[SArray->GetPos(DArray->GetPos(i+j+1)-1+pathdepth[x+j+1])] =
               pathdepth[x+j+1];
            }
            else if (D<256) {
               pathdepth[SArray->GetPos(DArray->GetPos(i+j+1)-1-pathdepth[x+j+1])] =
                 pathdepth[x+j+1]+ pathdepth[i+j+1] + 1;
               pathdepth[SArray->GetPos(DArray->GetPos(i+j+1)+pathdepth[i+j+1])] =
                 pathdepth[x+j+1]+ pathdepth[i+j+1] + 1;
            }
            DArray->SetB(i+j+1,1);
            DArray->SetPos(i+j+1,x+j+1);
            newsize--;
            j++;
         }
         i = i+j+1;
      }
   }
   for (i=0; i<n; i++) {
      j = SArray->GetPos(i);
      x = i;
      while (SArray->GetB(j)==0) {
         index = j;
         j = SArray->GetPos(index);
         SArray->SetPos(index,x);
         SArray->SetB(index,1);
         x = index;
      }
   }
   for (i=0; i<n; i++)
     assert(SArray->GetPos(i)<n);
   return newsize;
}

unsigned CalcNewSize(TSA *SArray)
{
   unsigned newsize=0;
   unsigned i=0;
   unsigned j;
   unsigned x;
   unsigned char *pathdepth=(unsigned char *)malloc(n*sizeof(unsigned char));
   TSA *DArray= new TSA();
   DArray->CreatePosArray(bitsn,n+1);
   unsigned index;
   newsize=n;
   for (i=0; i<n; i++) {
      DArray->SetPos(SArray->GetPos(i),i);
      DArray->SetB(i,0);
      SArray->SetB(i,0);
      pathdepth[i]=0;
   }
   for (i=0; i<n; i++) {
      x = DArray->GetPos(i);
      DArray->SetPos(i,SArray->GetPos(i));
      SArray->SetPos(i,x);
   }
   i = 0;
   for (i=1; i<n-1; i++) {
      if (DArray->GetPos(i) == 0)
         x = n+1;
      else
         x = SArray->GetPos(DArray->GetPos(i)-1);
      if (x < n-1 && pathdepth[i] + pathdepth[x] < D &&
          pathdepth[i+1] + pathdepth[x+1] < D && DArray->GetPos(i+1) > 0 &&
          SArray->GetPos(DArray->GetPos(i+1)-1) == x+1) {
         if (pathdepth[x] == 0) {
            pathdepth[x] = pathdepth[i]+1;
            pathdepth[SArray->GetPos(DArray->GetPos(i)-1+pathdepth[x])] = pathdepth[x];
         }
         else {
            pathdepth[SArray->GetPos(DArray->GetPos(i)-1-pathdepth[x])] =
              pathdepth[x]+ pathdepth[i] + 1;
            pathdepth[SArray->GetPos(DArray->GetPos(i)+pathdepth[i])] =
              pathdepth[x]+ pathdepth[i] + 1;
         }
         if (pathdepth[x+1] == 0) {
            pathdepth[x+1] = pathdepth[i+1]+1;
            pathdepth[SArray->GetPos(DArray->GetPos(i+1)-1+pathdepth[x+1])] =
            pathdepth[x+1];
         }
         else {
            pathdepth[SArray->GetPos(DArray->GetPos(i+1)-1-pathdepth[x+1])] =
              pathdepth[x+1]+ pathdepth[i+1] + 1;
            pathdepth[SArray->GetPos(DArray->GetPos(i+1)+pathdepth[i+1])] =
              pathdepth[x+1]+ pathdepth[i+1] + 1;
         }
         newsize--;
         j = 1;
         while (j+1 < C && x+j+1 <n && i+j+1 <n-1 &&
                pathdepth[i+j+1] + pathdepth[x+j+1] < D &&
                DArray->GetPos(i+j+1) > 0 &&
                SArray->GetPos(DArray->GetPos(i+j+1)-1) == x+j+1) {
            if (pathdepth[x+j+1] == 0) {
               pathdepth[x+j+1] = pathdepth[i+j+1]+1;
               pathdepth[SArray->GetPos(DArray->GetPos(i+j+1)-1+pathdepth[x+j+1])] =
               pathdepth[x+j+1];
            }
            else {
               pathdepth[SArray->GetPos(DArray->GetPos(i+j+1)-1-pathdepth[x+j+1])] =
                 pathdepth[x+j+1]+ pathdepth[i+j+1] + 1;
               pathdepth[SArray->GetPos(DArray->GetPos(i+j+1)+pathdepth[i+j+1])] =
                 pathdepth[x+j+1]+ pathdepth[i+j+1] + 1;
            }
            newsize--;
            j++;
         }
         i = i+j+1;
      }
   }
   for (i=0; i<n; i++) {
      j = SArray->GetPos(i);
      x = i;
      while (SArray->GetB(j)==0) {
         index = j;
         j = SArray->GetPos(index);
         SArray->SetPos(index,x);
         SArray->SetB(index,1);
         x = index;
      }
   }
   return newsize;
}

void CreateCompactedSuffixArray(TSA *SArray, TSA *DArray,int newsize)
{
   unsigned i;
   unsigned j=0;
   unsigned x=0;
   SArray->SetB(0,0);
   for (i=0;i<n;i++)
      if (SArray->GetB(i)==0) {
         DArray->SetPos(j,SArray->GetPos(i));
         DArray->SetB(j,0);
         SArray->SetPos(i,j++);
      }
      else {
         DArray->SetPos(j,SArray->GetPos(i));
         DArray->SetB(j,1);
         DArray->SetDepth(j,1);
         SArray->SetPos(i,j++);
         SArray->SetB(i++,0);
         while (SArray->GetB(i)) {
            x = DArray->GetDepth(j-1);
            SArray->SetPos(i++,x);
            DArray->SetDepth(j-1,x+1);
         }
         DArray->SetPos(j,SArray->GetPos(i));
         DArray->SetB(j,0);
         SArray->SetPos(i,j++);
         SArray->SetB(i,0);
      }
   for (i=0; i<newsize; i++)
         if (DArray->GetB(i) && SArray->GetB(DArray->GetPos(i)) == 0) {
            DArray->SetPos(i, SArray->GetPos(DArray->GetPos(i)));
            DArray->SetJump(i,0);
         }
         else if (DArray->GetB(i)) {
            DArray->SetJump(i,SArray->GetPos(DArray->GetPos(i)));
            DArray->SetPos(i,SArray->GetPos(DArray->GetPos(i)-DArray->GetJump(i)));
         }
}


void CopyAdjacentLCPsToCSA(TSA *SArray,TSA *CArray)
{
   unsigned i,j=0, minLCP;
   for (i=0; i<n; i++) {
      if (SArray->GetB(i) == 0)
         CArray->SetLCP(j++,minn(maxLCP,SArray->GetLCP(i)));
      else {
         j++;
         minLCP = SArray->GetLCP(i++);
         while (SArray->GetB(i)) {
            if (SArray->GetLCP(i) < minLCP)
               minLCP = SArray->GetLCP(i);
            i++;
         }
         if (SArray->GetLCP(i) < minLCP)
            CArray->SetLCP(j++,minn(maxLCP,SArray->GetLCP(i)));
         else
            CArray->SetLCP(j++,minn(maxLCP,minLCP));
      }
   }
}


void CompactSuffixArray(int alg)
{
   unsigned newsize=n;
   if (C==0) {
      unsigned i, minC=0;
      float minsize=n*(ceil(log(1.0*n)/log(2.0))+1)/8;
      for (i=1; i<16; i++) {
         C = (1 << i)-1;
         newsize = CalcNewSize(SA);
         printf("C=%d, newsize=%d\n", C, newsize);
         if (newsize*((ceil(log(1.0*n)/log(2.0))+1)/8+1.0*i/8) < minsize) {
            minsize = newsize*((ceil(log(1.0*n)/log(2.0))+1)/8+1.0*i/8);
            minC = C;
         }
      }
      C = minC;
      maxLCP = (C+1)*(C+1)/2-1;
   }
   TSA *SATemp = new TSA();
   SATemp->CreatePosArray(bitsn,n+1);
   newsize = MarkLinksToSuffixArray(SA, SATemp);
   if (!savemem && (alg == 2 || alg == 5)) {
      SA->CreateLCPArray(bitsn+1,n+1,false);
      CalcAdjacentLCPsSA(SA);
   }
   SA->DeletePosArray();
   CSA = new TSA();
   unsigned bitsd;
   if (alg == 5) {
      maxLCP = (1u << ((unsigned)ceil(log(1.0*C)/log(2.0))*2-1))-1;
      bitsd = (unsigned)ceil(log(1.0*maxx((C+1)*(C+1),(maxLCP+1)*2))/log(2.0));
   }
   else {
      bitsd = (unsigned)ceil(log(1.0*maxx((C+1),(maxLCP+1)*2))/log(2.0));
   }
   bool shortarray=true;
   if (alg==2 || alg==5)
      shortarray=false;
   CSA->CreateLCPArray(bitsd,newsize+1,shortarray);
   if (!savemem && (alg == 2 || alg == 5)) {
      SATemp->lcp = SA->lcp;
      SATemp->lcpbits = SA->lcpbits;
      CopyAdjacentLCPsToCSA(SATemp,CSA);
      SA->DeleteLCPArray();
      delete SA;
   }
   CSA->CreatePosArray(bitsn,newsize+1);
   SuffixArrayCompactSize = newsize;
   CreateCompactedSuffixArray(SATemp,CSA,newsize);
   SATemp->DeletePosArray();
   delete SATemp;
   if (!savemem && (alg == 2 || alg == 5))
      CalcLCPsFromAdjacentLCPs(CSA,0,newsize-1);
   else if (alg == 2 || alg == 5)
      CalcLCPs(CSA,0,newsize-1);
   printf("Compacted array takes %d*%f B = %f B space \n",
          newsize,(1.0*CSA->posbits+1)/8+
          1.0*CSA->lcpbits/8,
          newsize*((1.0*CSA->posbits+1)/8+1.0*CSA->lcpbits/8));
   fflush(stdout);
}

void UnCompactLink(TSA *CArray,unsigned Cindex, unsigned depth,
                   TSA *Area, unsigned &Areasize,
                   unsigned remaining, unsigned startparent)
{
   unsigned i=0;
   unsigned k=0;
   unsigned l;
   unsigned j=CArray->GetPos(Cindex);
   unsigned startthis = CArray->GetJump(Cindex)+startparent;
   unsigned thisdepth = minn(CArray->GetDepth(Cindex)-startparent,remaining);
   unsigned Areastart = Areasize;
   while (startthis > 0)
      if (!CArray->GetB(j+i)) {
         i++;
         startthis--;
         operations++;
      }
      else
         if (CArray->GetDepth(j+i) <= startthis)  {
            startthis -= CArray->GetDepth(j+i++);
            operations++;
         }
         else
            break;
   while (1) {
      if (CArray->GetB(j+i)) {
         UnCompactLink(CArray,j+i,depth+1,Area,Areasize,thisdepth-
         (Areasize-Areastart),startthis);
	 operations++;
         if (Areasize-Areastart>= thisdepth)
            break;
         startthis=0;
      }
      else {
         Area->SetPos(Areasize++,CArray->GetPos(j+i)+depth+1);
         sumdepth += depth+1;
	 operations++;
         if (Areasize-Areastart >= thisdepth)
            break;
      }
      i++;
   }
   if (depth+1 > maxdepth)
      maxdepth = depth+1;
}

void UnCompactSuffixArray()
{
   TSA *DArray = new TSA();
   DArray->CreatePosArray(bitsn,n+1);
   unsigned linkstart=0;
   unsigned newsize;
   unsigned i,j,k,dd, oldj;
   dd=1;
   j=0;
   maxdepth = 0;
   unsigned links=0;
   sumdepth = 0.0;
   unsigned suffixes=0;
   operations=0;
   for (i=0; i<10000; i++)
      depthcount[i]=0;
   for (i=0; i<SuffixArrayCompactSize;i++) {
      if (CSA->GetB(i)) {
         linkstart=0;
         oldj = j;
         links++;
         startownclock();
	 operations++;
         UnCompactLink(CSA,
                       i,
                       0,
                       DArray,
                       j,
                       CSA->GetDepth(i),
                       0);
         tuc = stopownclock();
         if (tuc > maxtuc) {
            maxtuc = tuc;
            maxtucindex = i;
            maxtucsize = j-oldj;
         }
	 if (maxdepth > D) { // !ERROR!
	    int t;
	    printf("Error. Link:\n");
	    for (t=oldj;t<j;t++)
	       printf("%d ",DArray->GetPos(t));
	    printf("\n");
	    maxdepth = D;
	 }
      }
      else {
         DArray->SetPos(j++,CSA->GetPos(i));
         suffixes++;
      }
      if (print)
         if (j*10.0/dd > n) {
            printf("%d ",10*dd++);
            fflush(stdout);
         }
   }
   double sumdepthv=0.0;
   for (i=0; i<10000; i++)
      sumdepthv +=depthcount[i]*(i-sumdepth/(n-suffixes))*(i-sumdepth/(n-suffixes));
   if (print)
      printf("\n");
   if (print)
      printf("Max uncompactime %4fs at index %d (jump=%d, length=%d),
      maxdepth=%d, avgdepth=%f, sqrt(variance)=%f, operations=%d, links=%d\n",
              maxtuc*1.0/hzz,maxtucindex,(int)(CSA->GetJump(maxtucindex)),maxtucsize,maxdepth,
              sumdepth/(n-suffixes),sqrt(sumdepthv/(n-suffixes)),operations,links);
   newsize = j;
   j=n+1;
   printf("UnCompacted array takes %d*%f B = %f B space \n",
          newsize,1.0*DArray->posbits/8,
          newsize*1.0*DArray->posbits/8);
   DArray->DeletePosArray();
   delete DArray;
}


void CreateSuffixArray()
{
   int *x, *p;
   p=(int*)malloc((n+1)*sizeof *p);
   x=(int*)malloc((n+1)*sizeof *x);
   unsigned i;
   for (i=0; i<n; i++)
      x[i]=(int)text[i]+128;
   x[n]=0;   
   suffixsort(x,p,n,256,0);
   free(x);
   unsigned k=0;
   while (((n+1)<<k)>>31 == 0) k++;
   bitsn = 32-k;
   SA = new TSA();
   SA->CreatePosArray(bitsn,n+1);
   SA->lcpbits = 0;
   for (i=0; i<n; i++) {
      SA->SetPos(i,p[i+1]);
      SA->SetB(i,0);
   }
   free(p);
}

void MakeSuffixArray(int alg)
{
   CreateSuffixArray();
   if (alg == 3) {
      SA->CreateLCPArray(bitsn+1,n+1,false);
      CalcAdjacentLCPsSA(SA);
      TSA * SATemp = new TSA();
      SATemp->CreateLCPArray((unsigned)ceil(log(maxLCP+1)/log(2.0))+1,n+1,false);
      unsigned i;
      for (i=0; i<n; i++)
         SATemp->SetLCP(i,SA->GetLCP(i));
      SA->DeleteLCPArray();
      SA->lcp = SATemp->lcp;
      SA->lcpbits = SATemp->lcpbits;
      delete SATemp;
   }
   else if (savemem && (alg == 2 || alg == 5)) {
      SA->CreateLCPArray(bitsn+1,n+1,false);
      CalcAdjacentLCPsSA(SA);
      SA->DeleteLCPArray();
   } else {
      maxLCP = 0;
      SA->lcpbits = 0;
   }
   printf("Array takes %d*%f B = %f B space \n",
          n,(1.0*SA->posbits+SA->lcpbits)/8,
          n*((1.0*SA->posbits+SA->lcpbits)/8));
   printf("Text size is %d B\n",n);
   if (alg == 0 || alg == 2 || alg == 4 || alg == 5)
     CompactSuffixArray(alg);
   else if (alg == 3)
     CalcLCPsFromAdjacentLCPs(SA,0,n-1);
}

void PrintSuffixArray(TSA *SArray, unsigned length)
{
  printf("length=%d\n",length);
   unsigned i;
   for (i=0; i<256 && i < length; i++) {
      if (!SArray->GetB(i) && SArray->lcpbits == 0)
         printf("%d:(%d) %s \n",i,SArray->GetPos(i),SubString(text,SArray->GetPos(i),20));
      else if (!SArray->GetB(i) && SArray->GetLCPl(i))
         printf("%d:(%d) %s (l=%d)\n",i,SArray->GetPos(i),SubString(text,SArray->GetPos(i),20),
                (int)SArray->GetLCP(i));
      else if (!SArray->GetB(i) && SArray->GetLCPr(i))
         printf("%d:(%d) %s (r=%d)\n",i,SArray->GetPos(i),SubString(text,SArray->GetPos(i),20),
                (int)SArray->GetLCP(i));
      else
         printf("%d:->%d (%d) (%d) \n",i,SArray->GetPos(i),
                (int)SArray->GetDepth(i),(int)SArray->GetJump(i));
      fflush(stdout);
   }
}


